#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sjisdec.h"
#include "sjismap.h"
#include "dfsys.h"

/*
  fileonhost -> fileondisk
  (hostfilelist, outdcpfile)
  File contains entries
    "path/on/filesys.ext" "path/on/disk.ext"
  Where the first path is path on the host filesystem, the second path is the path it will have on the DCP disk image.
  The DCP image is created with these files, then stored to the output file.
*/

unsigned int cvtu8tou16(char const* u8str, unsigned int u8len, wchar_t* u16str);

FILE* u8fopen(char const* path, char const* mode) {
  FILE* file = NULL;
#ifdef _WIN32
  wchar_t widemode[128];
  unsigned int modelen = strlen(mode);
  modelen = modelen >= 128 ? 127 : modelen;
  unsigned int widelen = cvtu8tou16(mode, modelen, widemode);
  widemode[widelen] = 0;
  unsigned int u8len = strlen(path);
  wchar_t* u16path = malloc(sizeof(wchar_t) * (u8len + 1));
  unsigned int u16len = cvtu8tou16(path, u8len, u16path);
  u16path[u16len] = 0;
  file = _wfopen(u16path, widemode);
  free(u16path);
#else
  file = fopen(path, mode);
#endif
  return file;
}

unsigned int cvtu8tou16(char const* u8str, unsigned int u8len, wchar_t* u16str) {
  unsigned int wlen = 0;
  unsigned int cpoint = 0;
  unsigned int remlen = 0;
  for (unsigned int idx = 0; idx < u8len; ++idx) {
    unsigned int cc = (unsigned int)u8str[idx] & 0xff;
    switch (cc >> 5) {
      case 0: case 1: case 2: case 3:
        remlen = 0;
        u16str[wlen++] = (wchar_t)cc;
        break;
      case 4: case 5:
        cpoint = cpoint << 6 | (cc & 0x3f);
        if (remlen == 1) { // write codepoint
          if (cpoint < 0x10000) { // 0xd800 - 0xdfff ?
            u16str[wlen++] = cpoint;
          } else {
            u16str[wlen] = (0xd800 - 0x40) + (cpoint >> 10);
            u16str[wlen + 1] = 0xdc00 + (cpoint & 0x3ff);
            wlen += 2;
          }
        }
        remlen -= (remlen != 0);
        break;
      case 6:
        remlen = 1;
        cpoint = cc & 0x1f;
        break;
      case 7:
        remlen = (0x1fffffe >> (cc & 0x1f) & 2) | (0xff0000 >> (cc & 0x1f) & 1);
        cpoint = cc & (0x3f >> remlen);
        break;
    }
  }
  return wlen;
}

char const fmchtypes[] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 0, 0, 0, 0, 0, 0,
  0, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 4,
  0, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

unsigned char fmishexl(char const* str, unsigned int len) {
  unsigned char flags = 3;
  for (unsigned int idx = 0; idx < len; ++idx) {
    flags &= fmchtypes[(unsigned int)str[idx] & 0xff];
  }
  return (len != 0) & (flags & 1);
}

unsigned char fmisdecl(char const* str, unsigned int len) {
  unsigned char flags = 3;
  for (unsigned int idx = 0; idx < len; ++idx) {
    flags &= fmchtypes[(unsigned int)str[idx] & 0xff];
  }
  return (len != 0) & (flags >> 1 & 1);
}

unsigned int fmfromhexl(char const* str, unsigned int len) {
  unsigned int result = 0;
  for (unsigned int idx = 0; idx < len; ++idx) {
    char c = str[idx];
    result = result << 4 | ((c & 0x0f) + (c >> 6) * 9);
  }
  return result;
}

unsigned int fmfromdecl(char const* str, unsigned int len) {
  unsigned int result = 0;
  for (unsigned int idx = 0; idx < len; ++idx) {
    char c = str[idx];
    result = result * 10 + (c & 0x0f);
  }
  return result;
}

unsigned int decodenuml(char const* str, unsigned int slen) {
  unsigned int result = (unsigned int)-1;
  if (slen > 2 && str[0] == '0' && (str[1] == 'x' || str[1] == 'X') && fmishexl(str + 2, slen - 2)) {
    result = fmfromhexl(str + 2, slen - 2);
  } else if (slen > 1 && (str[slen - 1] == 'h' || str[slen - 1] == 'H') && fmishexl(str, slen - 1)) {
    result = fmfromhexl(str, slen - 1);
  } else if (fmisdecl(str, slen)) {
    result = fmfromdecl(str, slen);
  } else {
    char* tempb = malloc(slen + 1);
    memcpy(tempb, str, slen);
    tempb[slen] = '\0';
    printf("Bad number to decode %s\n", tempb);
    fflush(stdout);
    free(tempb);
  }
  return result;
}

struct strbuffer {
  char* strdata;
  unsigned int* entries;
  unsigned int entrynum;
  unsigned int datasize;
  unsigned int entrycap;
  unsigned int datacap;
};

void strbufffree(struct strbuffer* sbuff) {
  free(sbuff->entries);
  free(sbuff->strdata);
  sbuff->strdata = NULL;
  sbuff->entries = NULL;
  sbuff->entrynum = 0;
  sbuff->datasize = 0;
  sbuff->entrycap = 0;
  sbuff->datacap = 0;
}

unsigned char strbuffinit(struct strbuffer* sbuff, unsigned int datacap, unsigned int entrycap) {
  unsigned char result = 1;
  sbuff->strdata = malloc(datacap);
  sbuff->entries = sbuff->strdata != NULL ? malloc(sizeof(unsigned int) * entrycap * 2) : NULL;
  if (sbuff->entries != NULL) {
    sbuff->entrynum = 0;
    sbuff->datasize = 0;
    sbuff->entrycap = entrycap;
    sbuff->datacap = datacap;
  } else {
    strbufffree(sbuff);
    result = 0;
  }
  return result;
}

unsigned char strbuffadd(struct strbuffer* sbuff, char const* str, unsigned int slen) {
  unsigned char result = 1;
  if (sbuff->datasize + slen > sbuff->datacap) {
    unsigned int dcap2 = sbuff->datacap == 0 ? 256 : sbuff->datacap * 2;
    while (sbuff->datasize + slen > dcap2) {
      dcap2 *= 2;
    }
    char* newdata = realloc(sbuff->strdata, dcap2);
    if (newdata != NULL) {
      sbuff->strdata = newdata;
      sbuff->datacap = dcap2;
    } else {
      result = 0;
    }
  }
  if (result && sbuff->entrynum >= sbuff->entrycap) {
    unsigned int* newentries = realloc(sbuff->entries, sizeof(unsigned int) * (sbuff->entrycap * 4));
    if (newentries != NULL) {
      sbuff->entries = newentries;
      sbuff->entrycap = sbuff->entrycap * 2;
    } else {
      result = 0;
    }
  }
  if (result) {
    sbuff->entries[sbuff->entrynum * 2] = sbuff->datasize;
    sbuff->entries[sbuff->entrynum * 2 + 1] = slen;
    ++sbuff->entrynum;
    memcpy(sbuff->strdata + sbuff->datasize, str, slen);
    sbuff->datasize += slen;
  }
  return result;
}

void capitals(char* str, unsigned int slen) {
  for (unsigned int idx = 0; idx < slen; ++idx) {
    char c = str[idx];
    str[idx] -= ((c >= 'a') & (c <= 'z')) << 5;
  }
}

unsigned char mmreadline(FILE* file, char* out, unsigned int* slen) {
  unsigned char eol = 0;
  unsigned int len = 0;
  char c;
  unsigned int maxlen = *slen;
  while (len < maxlen && (c = getc(file)) != EOF && !(eol = (c == '\n'))) {
    out[len++] = c;
  }
  if (len < maxlen) {
    if (len > 0 && out[len - 1] == '\r') {
      --len;
    }
    out[len] = 0;
  }
  *slen = len;
  return eol;
}

void strskipws(unsigned int* intsidx, char const* line, unsigned int rlen) {
  char c;
  while (*intsidx < rlen && ((c = line[*intsidx]) == ' ' || c == '\t')) {
    ++*intsidx;
  }
}

void strskipnonws(unsigned int* intsidx, char const* line, unsigned int rlen) {
  char c;
  while (*intsidx < rlen && (c = line[*intsidx]) != ' ' && c != '\t') {
    ++*intsidx;
  }
}

void strskipnondq(unsigned int* intsidx, char const* line, unsigned int rlen) {
  while (*intsidx < rlen && line[*intsidx] != '"') {
    ++*intsidx;
  }
}

// return number of token (start, length) pairs
unsigned int strtokens(char const* line, unsigned int slen, unsigned int* tokenbounds, unsigned int maxtokens) {
  unsigned int numtokens = 0;
  unsigned int sidx = 0;
  unsigned char indq = 0;
  unsigned int tokenstart = 0;
  while (sidx < slen && numtokens < maxtokens) {
    unsigned int cc = line[sidx];
    if (indq) {
      if (cc == '"') {
        indq = 0;
        tokenbounds[numtokens * 2] = tokenstart;
        tokenbounds[numtokens * 2 + 1] = sidx - tokenstart;
        ++numtokens;
        tokenstart = sidx + 1;
      }
    } else {
      indq = (cc == '"');
      if (cc == ' ' || cc == '\t' || cc == '"') {
        if (tokenstart < sidx) {
          tokenbounds[numtokens * 2] = tokenstart;
          tokenbounds[numtokens * 2 + 1] = sidx - tokenstart;
          ++numtokens;
        }
        tokenstart = sidx + 1;
      } else {
        //
      }
    }
    ++sidx;
  }
  if (numtokens < maxtokens && !indq && tokenstart < sidx) {
    tokenbounds[numtokens * 2] = tokenstart;
    tokenbounds[numtokens * 2 + 1] = sidx - tokenstart;
    ++numtokens;
  }
  return numtokens;
}

struct dextradata {
  unsigned char flags;
  unsigned int creationtime;
  unsigned short clusteroffset;
  unsigned int datalength;
  unsigned int hostpath;
};

// set offsets
unsigned int dcpfssetoffsets(struct fsdir* parentdir, unsigned int clusterstart, unsigned int clustersize, char* reservedclusters, unsigned char filepass, struct dextradata* extradata) {
  unsigned int currentofs = clusterstart;
  unsigned int* sortidx = malloc(sizeof(unsigned int) * (filepass ? parentdir->dirnum + parentdir->filenum : parentdir->dirnum));
  unsigned int snum = 0;
  unsigned int fnum = 0;
  for (unsigned int didx = 0; didx < parentdir->dircap; ++didx) {
      struct fsfile const* file = &parentdir->dirs[didx].filedata;
      if (file->keyname != (unsigned int)-1) {
        sortidx[snum] = didx;
        ++snum;
      }
    }
  if (filepass) {
    for (unsigned int fidx = 0; fidx < parentdir->filecap; ++fidx) {
      struct fsfile const* file = &parentdir->files[fidx];
      if (file->keyname != (unsigned int)-1) {
        sortidx[snum + fnum] = fidx; // parentdir->dircap + 
        ++fnum;
      }
    }
  }
  {
    unsigned int sidx = 0;
    while (sidx + 1 < snum) {
      if (parentdir->dirs[sortidx[sidx + 1]].filedata.fileidx > parentdir->dirs[sortidx[sidx]].filedata.fileidx) { // strcmp(knames[sortidx[sidx]], knames[sortidx[sidx + 1]])
        unsigned int temp = sortidx[sidx];
        sortidx[sidx] = sortidx[sidx + 1];
        sortidx[sidx + 1] = temp;
        sidx -= (sidx != 0);
      } else {
        ++sidx;
      }
    }
    if (filepass) {
      sidx = 0;
      unsigned int* sortidx2 = sortidx + snum;
      while (sidx + 1 < fnum) {
        if (parentdir->files[sortidx2[sidx + 1]].fileidx > parentdir->files[sortidx2[sidx]].fileidx) {
          unsigned int temp = sortidx2[sidx];
          sortidx2[sidx] = sortidx2[sidx + 1];
          sortidx2[sidx + 1] = temp;
          sidx -= (sidx != 0);
        } else {
          ++sidx;
        }
      }
    }
  }
  if (filepass) {
    unsigned int const* sortidx2 = sortidx + snum;
    for (unsigned int fidx = 0; fidx < fnum; ++fidx) {
      struct fsfile* file = &parentdir->files[sortidx2[fidx]]; //  - parentdir->dircap
      struct dextradata* dexd = &extradata[file->fileidx >> 1];
      if (dexd->clusteroffset == (unsigned short)-1) {
        unsigned int numclusters = (dexd->datalength + clustersize - 1) / clustersize;
        numclusters += (numclusters == 0);
        unsigned int urlen = 0;
        while (urlen < numclusters) {
          urlen = reservedclusters[currentofs++] ? 0 : urlen + 1;
        }
        dexd->clusteroffset = currentofs - numclusters;
      }
    }
  } else if (parentdir->filedata.parentdir != NULL) { // not root dir
    struct dextradata* dexd = &extradata[parentdir->filedata.fileidx >> 1];
    if (dexd->clusteroffset == (unsigned short)-1) {
      unsigned int datasize = (parentdir->dirnum + parentdir->filenum + 2) * 32 + 32; // numdirentries != 0 ? numdirentries * 32 : 
      unsigned int numclusters = (datasize + clustersize - 1) / clustersize;
      numclusters += (numclusters == 0);
      unsigned int urlen = 0;
      while (urlen < numclusters) {
        urlen = reservedclusters[currentofs++] ? 0 : urlen + 1;
      }
      dexd->clusteroffset = currentofs - numclusters;
      memset(reservedclusters + (currentofs - numclusters), 1, numclusters);
    }
  }
  for (unsigned int sidx = 0; sidx < snum; ++sidx) {
    currentofs = dcpfssetoffsets(&parentdir->dirs[sortidx[sidx]], currentofs, clustersize, reservedclusters, filepass, extradata);
  }
  //free(knames);
  free(sortidx);
  return currentofs;
}

void dcpmarkreserved(struct fsdir* parentdir, unsigned int clustersize, char* reservedclusters, struct dextradata* extradata) {
  for (unsigned int fidx = 0; fidx < parentdir->filecap; ++fidx) {
    struct fsfile* file = &parentdir->files[fidx];
    struct dextradata* dexd = &extradata[file->fileidx >> 1];
    if (file->keyname != (unsigned int)-1 && dexd->clusteroffset != (unsigned short)-1) {
      unsigned int numclusters = (dexd->datalength + clustersize - 1) / clustersize;
      memset(reservedclusters + dexd->clusteroffset, 1, numclusters);
    }
  }
  for (unsigned int didx = 0; didx < parentdir->dircap; ++didx) {
    struct fsdir* dir = &parentdir->dirs[didx];
    if (dir->filedata.keyname != (unsigned int)-1) {
      struct dextradata* dexd = &extradata[dir->filedata.fileidx >> 1];
      if (dexd->clusteroffset != (unsigned short)-1) {
        unsigned int datasize = (parentdir->dirnum + parentdir->filenum + 2) * 32 + 32;
        unsigned int numclusters = (datasize + clustersize - 1) / clustersize;
        memset(reservedclusters + dexd->clusteroffset, 1, numclusters);
      }
      dcpmarkreserved(dir, clustersize, reservedclusters, extradata);
    }
  }
}

struct fsfile* dcpaddfile(struct fsfile* parentdir, char const* diskpath, struct idxalloc* ial) {
  char keyname[256];
  char filename[256];
  unsigned int slashofs = 0;
  while (diskpath[slashofs] != 0 && diskpath[slashofs] != '/') {
    ++slashofs;
  }
  struct fsfile* rfile = NULL;
  if (!(parentdir->fileidx & 1) && slashofs > 0) {
    memcpy(filename, diskpath, slashofs);
    capitals(filename, slashofs);
    unsigned int sjislen = convertutf8tosjis(filename, slashofs, keyname, cvtsjisvalues, cvtsjislookuptable);
    unsigned int sjislen2 = convertutf8tosjis(diskpath, slashofs, filename, cvtsjisvalues, cvtsjislookuptable);
    keyname[sjislen] = 0;
    filename[sjislen2] = 0;
    if (diskpath[slashofs] == '/') { // add dir
      struct fsfile* subfile = fsgetfile((struct fsdir*)parentdir, keyname);
      struct fsfile* subdir = NULL;
      if (subfile != NULL) {
        subdir = (subfile->fileidx & 1) ? NULL : subfile;
      } else {
        subdir = (struct fsfile*)fsadddirectory((struct fsdir*)parentdir, filename, keyname, ial);
      }
      if (subdir != NULL) {
        rfile = dcpaddfile(subdir, diskpath + slashofs + 1, ial);
      }
    } else { // add and return file
      rfile = fsaddfile((struct fsdir*)parentdir, filename, keyname, ial);
    }
  }
  return rfile;
}

struct diskparameters {
  unsigned char disktype;
  unsigned char tracks[160];
  unsigned char trackflag;

  unsigned char version;
  // 2.0 bios param block (13)
  unsigned short bytespersector;
  unsigned char sectorspercluster;
  unsigned short reservedsectors;
  unsigned char numfats;
  unsigned short maxrootdirentries;
  unsigned short numsectors16;
  unsigned char mediadesc;
  unsigned short sectorsperfat;
  // 3.0 (6)
  unsigned short sectorspertrack;
  unsigned short numheads;
  unsigned short hiddensectors;
};

struct fsfile const** getallfiles(struct fsdir const* parentdir, struct fsfile const** list, unsigned int* listofs, unsigned int* listmax) {
  unsigned int flistofs = *listofs;
  unsigned int expsize = flistofs + parentdir->dirnum + parentdir->filenum;
  if (expsize > *listmax) {
    unsigned int newcap = *listmax * 2;
    while (expsize > newcap) {
      newcap *= 2;
    }
    struct fsfile const** list2 = (struct fsfile const**)realloc(list, sizeof(struct fsfile const*) * newcap);
    if (list2) {
      list = list2;
      *listmax = newcap;
    }
  }
  for (unsigned int didx = 0; didx < parentdir->dircap; ++didx) {
    struct fsfile* file = &parentdir->dirs[didx].filedata;
    if (file->keyname != (unsigned int)-1) {
      list[flistofs++] = file;
    }
  }
  for (unsigned int fidx = 0; fidx < parentdir->filecap; ++fidx) {
    struct fsfile* file = &parentdir->files[fidx];
    if (file->keyname != (unsigned int)-1) {
      list[flistofs++] = file;
    }
  }
  *listofs = flistofs;
  return list;
}

unsigned int writebuffered(FILE* diskfile, char const* data, unsigned int datasize, char* buffer, unsigned int buffersize, unsigned int granularity) {
  unsigned int dataofs = 0;
  while (dataofs + buffersize <= datasize) {
    fwrite(data + dataofs, 1, buffersize, diskfile);
    dataofs += buffersize;
  }
  unsigned int numgra = dataofs / granularity;
  if (dataofs < datasize) {
    unsigned int rem = datasize - dataofs;
    memcpy(buffer, data + dataofs, rem);
    unsigned int granulated = (rem + granularity - 1) / granularity;
    numgra += granulated;
    granulated *= granularity;
    memset(buffer + rem, 0, granulated - rem);
    fwrite(buffer, 1, granulated, diskfile);
    
  }
  return numgra;
}

unsigned int copyfilebuffered(FILE* dstfile, FILE* srcfile, char* buffer, unsigned int buffersize, unsigned int granularity) {
  unsigned int readamt = 0;
  unsigned int numgra = 0;
  unsigned int spc = buffersize / granularity;
  do {
    readamt = fread(buffer, 1, buffersize, srcfile);
    if (readamt == buffersize) {
      fwrite(buffer, 1, buffersize, dstfile);
      numgra += spc;
    }
  } while (readamt == buffersize);
  if (readamt > 0 && readamt < buffersize) {
    unsigned int granulated = (readamt + granularity - 1) / granularity;
    numgra += granulated;
    granulated *= granularity;
    memset(buffer + readamt, 0, granulated - readamt);
    fwrite(buffer, 1, granulated, dstfile);
  }
  return numgra;
}

unsigned int sgn2(int v) {
  return (unsigned int)(1 + (v > 0) - (v < 0));
}

void setu32(char* buffer, unsigned int val) {
  buffer[0] = (char)(val & 0xff);
  buffer[1] = (char)(val >> 8 & 0xff);
  buffer[2] = (char)(val >> 16 & 0xff);
  buffer[3] = (char)(val >> 24 & 0xff);
}

void setu16(char* buffer, unsigned int val) {
  buffer[0] = (char)(val & 0xff);
  buffer[1] = (char)(val >> 8 & 0xff);
}

void writeentry(char* buffer, char const* name, char const* ext, struct dextradata const* extradata) {
  unsigned int namelen = strlen(name);
  unsigned int extlen = strlen(ext);
  memset(buffer, ' ', 8 + 3);
  namelen = namelen > 8 ? 8 : namelen;
  extlen = extlen > 3 ? 3 : extlen;
  memcpy(buffer, name, namelen);
  memcpy(buffer + 8, ext, extlen);
  buffer[11] = (char)extradata->flags;
  setu32(buffer + 22, extradata->creationtime);
  setu16(buffer + 26, extradata->clusteroffset);
  setu32(buffer + 28, extradata->datalength);
}

void sepnamext(char const* str, char* namebuff, unsigned int maxnamelen, char* extbuff, unsigned int maxextlen) {
  unsigned int slen = strlen(str);
  unsigned int idx = slen;
  while (idx > 0 && str[idx] != '.') {
    --idx;
  }
  unsigned int namelen = 0,
               extlen = 0;
  if (str[idx] == '.') {
    namelen = idx;
    extlen = slen - (idx + 1);
    namelen = namelen > maxnamelen ? maxnamelen : namelen;
    extlen = extlen > maxextlen ? maxextlen : extlen;
  } else {
    namelen = slen;
    namelen = namelen > maxnamelen ? maxnamelen : namelen;
  }
  memcpy(namebuff, str, namelen);
  if (extlen) {
    memcpy(extbuff, str + idx + 1, extlen);
  }
  namebuff[namelen] = 0;
  extbuff[extlen] = 0;
}

unsigned int writebuffereddirentries(FILE* diskfile, struct fsdir const* dir, struct dextradata const* extradata, char* buffer, unsigned int buffersize, unsigned int granularity, int padding, unsigned char subdir) {
  unsigned int numentries = dir->dirnum + dir->filenum;
  unsigned int spc = buffersize / granularity;
  unsigned int numsectors = 0;

  struct fsfile** filearray = malloc(sizeof(struct fsfile*) * numentries);
  char const** knamearray = malloc(sizeof(char const*) * numentries);
  {
    unsigned int entryofs = 0;
    for (unsigned int didx = 0; didx < dir->dircap; ++didx) {
      struct fsfile* file = &dir->dirs[didx].filedata;
      if (file->keyname != (unsigned int)-1) {
        filearray[entryofs] = file;
        knamearray[entryofs] = fsgetkeyname(file);
        ++entryofs;
      }
    }
    for (unsigned int fidx = 0; fidx < dir->filecap; ++fidx) {
      struct fsfile* file = &dir->files[fidx];
      if (file->keyname != (unsigned int)-1) {
        filearray[entryofs] = file;
        knamearray[entryofs] = fsgetkeyname(file);
        ++entryofs;
      }
    }
  }
  unsigned int* sortidx = malloc(sizeof(unsigned int) * numentries);
  for (unsigned int sidx = 0; sidx < numentries; ++sidx) {
    sortidx[sidx] = sidx;
  }  
  {
    unsigned int sidx = 0;
    while (sidx + 1 < numentries) {
      unsigned int id1 = filearray[sortidx[sidx]]->fileidx;
      unsigned int id2 = filearray[sortidx[sidx + 1]]->fileidx;
      struct dextradata const* dexd1 = &extradata[id1>>1];
      struct dextradata const* dexd2 = &extradata[id2>>1];
      unsigned int lex1 = sgn2(((int)dexd1->flags & 0x04) - ((int)dexd2->flags & 0x04)) << 2 | sgn2(((int)id1 & 1) - ((int)id2 & 1));
      if (lex1 < 5 || (lex1 == 5 && strcmp(knamearray[sortidx[sidx]], knamearray[sortidx[sidx + 1]]) > 0)) {
        unsigned int temp = sortidx[sidx];
        sortidx[sidx] = sortidx[sidx + 1];
        sortidx[sidx + 1] = temp;
        sidx -= (sidx != 0);
      } else {
        ++sidx;
      }
    }
  }

  unsigned int buffofs = 0;
  // write subdir entries
  if (subdir) {
    writeentry(buffer, ".", "", &extradata[dir->filedata.fileidx>>1]);
    writeentry(buffer + 32, "..", "", &extradata[dir->filedata.parentdir->filedata.fileidx>>1]);
    buffofs += 64;
  }

  char cname[9];
  char cext[4];

  for (unsigned int eidx = 0; eidx < numentries; ++eidx) {
    struct fsfile const* file = filearray[sortidx[eidx]];
    char const* fname = fsgetfilename(file);
    sepnamext(fname, cname, 8, cext, 3);
    writeentry(buffer + buffofs, cname, cext, &extradata[file->fileidx>>1]);
    buffofs += 32;
    if (buffofs == buffersize) {
      fwrite(buffer, 1, buffersize, diskfile);
      numsectors += spc;
      buffofs = 0;
    }
  }
  memset(buffer + buffofs, padding, 32);
  buffer[buffofs] = 0;
  buffofs += 32;
  if (buffofs == buffersize) {
    fwrite(buffer, 1, buffersize, diskfile);
    numsectors += spc;
    buffofs = 0;
  }
  if (buffofs > 0) {
    unsigned int granulated = (buffofs + granularity - 1) / granularity;
    numsectors += granulated;
    granulated *= granularity;
    if (buffofs < granulated) {
      memset(buffer + buffofs, padding, granulated - buffofs);
    }
    for (unsigned int idx = buffofs; idx < granulated; idx += 32) {
      buffer[idx] = 0;
    }
    fwrite(buffer, 1, granulated, diskfile);
  }
  free(sortidx);
  free(filearray);

  return numsectors;
}

void writeu16(char* data, unsigned int val) {
  data[0] = (char)(val & 0xff);
  data[1] = (char)(val >> 8 & 0xff);
}

void dcpwritedisk(FILE* diskfile, struct diskparameters const* dpara, struct fsdir const* diskroot, struct strbuffer const* hostpaths, char const* codestub, unsigned int codelen, struct dextradata const* extradata) {
  unsigned short clusterlist[4096];
  char clusterlistbytes[6144];

  unsigned int sectorsize = dpara->bytespersector;
  unsigned int clustersize = dpara->bytespersector * dpara->sectorspercluster;
  char* buffer = (char*)malloc(clustersize);

  if (buffer) {
    buffer[0] = (char)dpara->disktype;
    memcpy(buffer + 1, dpara->tracks, 160);
    buffer[161] = (char)dpara->trackflag;
    fwrite(buffer, 1, 162, diskfile);

    unsigned int codewritten = codelen > clustersize ? clustersize : codelen;
    memcpy(buffer, codestub, codewritten);
    unsigned int codesectorbytes = ((codewritten + sectorsize - 1) / sectorsize + (codewritten == 0)) * sectorsize;
    if (codewritten < codesectorbytes) {
      memset(buffer + codewritten, 0, codesectorbytes - codewritten);
    }

    writeu16(buffer + 11, dpara->bytespersector);
    buffer[13] = (char)dpara->sectorspercluster;
    writeu16(buffer + 14, dpara->reservedsectors);
    buffer[16] = (char)dpara->numfats;
    writeu16(buffer + 17, dpara->maxrootdirentries);
    writeu16(buffer + 19, dpara->numsectors16);
    buffer[21] = (char)dpara->mediadesc;
    writeu16(buffer + 22, dpara->sectorsperfat);

    if (dpara->version >= 30) {
      writeu16(buffer + 24, dpara->sectorspertrack);
      writeu16(buffer + 26, dpara->numheads);
      writeu16(buffer + 28, dpara->hiddensectors);
    }
    fwrite(buffer, 1, codesectorbytes, diskfile);
    if (codewritten < codelen) {
      writebuffered(diskfile, buffer, codelen - codewritten, buffer, clustersize, sectorsize);
    }

    memset(clusterlist, 0, sizeof(unsigned short) * 4096);

    clusterlist[0] = 0xf00 | dpara->mediadesc;
    clusterlist[1] = 0xfff;

    struct fsfile const** filelist = malloc(sizeof(struct fsfile*) * 4096);
    unsigned int numfiles = 0;
    unsigned int listmax = 4096;
    filelist = getallfiles(diskroot, filelist, &numfiles, &listmax);
    unsigned int* sortidx = malloc(sizeof(unsigned int) * numfiles);

    for (unsigned int idx = 0; idx < numfiles; ++idx) {
      sortidx[idx] = idx;
    }

    unsigned int sidx = 0;
    while (sidx + 1 < numfiles) {
      struct dextradata const* dexd1 = &extradata[filelist[sortidx[sidx]]->fileidx>>1];
      struct dextradata const* dexd2 = &extradata[filelist[sortidx[sidx + 1]]->fileidx>>1];
      if (dexd1->clusteroffset > dexd2->clusteroffset) {
        unsigned int temp = sortidx[sidx];
        sortidx[sidx] = sortidx[sidx + 1];
        sortidx[sidx + 1] = temp;
        sidx -= (sidx != 0);
      } else {
        ++sidx;
      }
    }

    for (unsigned int fidx = 0; fidx < numfiles; ++fidx) {
      unsigned int numcluster = 0;
      struct fsfile const* file = filelist[sortidx[fidx]];
      struct dextradata const* dexd = &extradata[file->fileidx>>1];
      if ((file->fileidx & 1)) {
        numcluster = ((dexd->datalength + clustersize - 1) / clustersize);
      } else {
        struct fsdir const* asdir = (struct fsdir const*)((void*)file - offsetof(struct fsdir, filedata));
        numcluster = ((asdir->filenum + asdir->dirnum + 3) * 32 + clustersize - 1) / clustersize;
      }
      numcluster += (numcluster == 0);
      unsigned int firstcluster = dexd->clusteroffset;
      if (firstcluster >= 2 && firstcluster + numcluster <= 0xff0) {
        unsigned int clusterofs = 0;
        while (clusterofs + 1 < numcluster) {
          clusterlist[firstcluster + clusterofs] = firstcluster + clusterofs + 1;
          ++clusterofs;
        }
        clusterlist[firstcluster + numcluster - 1] = 0xfff;
      }
    }

    unsigned int cbofs = 0;
    for (unsigned int cidx = 0; cidx < 4096; cidx += 2) {
      unsigned int v = (clusterlist[cidx] & 0x0fff) | (clusterlist[cidx + 1] & 0x0fff) << 12;
      clusterlistbytes[cbofs] = (char)(v & 0xff);
      clusterlistbytes[cbofs + 1] = (char)(v >> 8 & 0xff);
      clusterlistbytes[cbofs + 2] = (char)(v >> 16 & 0xff);
      cbofs += 3;
    }

    for (unsigned int fidx = 0; fidx < dpara->numfats; ++fidx) {
      writebuffered(diskfile, clusterlistbytes, dpara->sectorsperfat * sectorsize, buffer, clustersize, sectorsize);
    }

    unsigned int numrootdirsectors = writebuffereddirentries(diskfile, diskroot, extradata, buffer, clustersize, sectorsize, 0xe5, 0);
    unsigned int targetrootdirsectors = (dpara->maxrootdirentries * 32 + sectorsize - 1) / sectorsize;

    if (numrootdirsectors < targetrootdirsectors) {
      memset(buffer, 0xe5, sectorsize);
      for (unsigned int idx = 0; idx < sectorsize; idx += 32) {
        buffer[idx] = 0;
      }
      while (numrootdirsectors < targetrootdirsectors) {
        fwrite(buffer, 1, sectorsize, diskfile);
        ++numrootdirsectors;
      }
    }

    unsigned int currentcluster = 2;
    for (unsigned int fidx = 0; fidx < numfiles; ++fidx) {
      unsigned int numcluster = 0;
      struct fsfile const* file = filelist[sortidx[fidx]];
      struct dextradata const* dexd = &extradata[file->fileidx>>1];
      unsigned int numsectors = 0;
      if (dexd->clusteroffset >= currentcluster) {
        if (dexd->clusteroffset > currentcluster) {
          memset(buffer, 0xe5, clustersize);
          while (currentcluster < dexd->clusteroffset) {
            fwrite(buffer, 1, clustersize, diskfile);
            ++currentcluster;
          }
        }
        unsigned int spc = dpara->sectorspercluster;
        if ((file->fileidx & 1)) {
          FILE* srcfile = u8fopen(hostpaths->strdata + hostpaths->entries[dexd->hostpath * 2], "rb");
          numsectors = copyfilebuffered(diskfile, srcfile, buffer, clustersize, sectorsize);
          fclose(srcfile);
          numcluster = (dexd->datalength + clustersize - 1) / clustersize;
          numcluster += (numcluster == 0);
        } else {
          struct fsdir const* asdir = (struct fsdir const*)((void*)file - offsetof(struct fsdir, filedata));
          numcluster = ((asdir->filenum + asdir->dirnum + 3) * 32 + clustersize - 1) / clustersize;
          numcluster += 1;
          numsectors = writebuffereddirentries(diskfile, asdir, extradata, buffer, clustersize, sectorsize, 0xe5, 1);
        }
        if (numsectors < numcluster * spc) {
          memset(buffer, 0xe5, (numsectors + spc <= numcluster * spc ? clustersize : sectorsize * (numcluster * spc - numsectors)));
          while (numsectors + spc <= numcluster * spc) {
            fwrite(buffer, 1, clustersize, diskfile);
            numsectors += spc;
          }
          if (numsectors < numcluster * spc) {
            fwrite(buffer, 1, sectorsize * (numcluster * spc - numsectors), diskfile);
          }
        }
        currentcluster += numcluster;
      }
    }

    free(sortidx);
    free(filelist);
    free(buffer);
  }
}

unsigned char cvtflags(char const* cstr, unsigned int len) {
  unsigned char result = 0;
  for (unsigned int idx = 0; idx < len; ++idx) {
    char cc = cstr[idx];
    result |= (unsigned char)((cc == '?') << 7 | (cc == 'I') << 6 | (cc == 'A') << 5 | (cc == 'D') << 4
                            | (cc == 'V') << 3 | (cc == 'S') << 2 | (cc == 'H') << 1 | (cc == 'R') << 0);
  }
  return result;
}

unsigned int datetotstamp(char const* date, unsigned int dlen, char const* time, unsigned int tlen) {
  unsigned int sep1 = 0;
  unsigned int sep2 = 0;
  unsigned int sep3 = 0;
  unsigned int sep4 = 0;
  while (sep1 < dlen && date[sep1] != '-') {
    ++sep1;
  }
  sep2 = sep1 + 1;
  while (sep2 < dlen && date[sep2] != '-') {
    ++sep2;
  }
  while (sep3 < tlen && time[sep3] != ':') {
    ++sep3;
  }
  sep4 = sep3 + 1;
  while (sep4 < tlen && time[sep4] != ':') {
    ++sep4;
  }
  unsigned int result = 0;
  if (sep2 + 1 < dlen && sep4 + 1 < tlen
        && fmisdecl(date, sep1) && fmisdecl(date + (sep1 + 1), sep2 - (sep1 + 1)) && fmisdecl(date + (sep2 + 1), dlen - (sep2 + 1))
        && fmisdecl(time, sep3) && fmisdecl(time + (sep3 + 1), sep4 - (sep3 + 1)) && fmisdecl(time + (sep4 + 1), tlen - (sep4 + 1))) {
    unsigned int year = fmfromdecl(date, sep1);
    unsigned int month = fmfromdecl(date + (sep1 + 1), sep2 - (sep1 + 1));
    unsigned int day = fmfromdecl(date + (sep2 + 1), dlen - (sep2 + 1));
    unsigned int hour = fmfromdecl(time, sep3);
    unsigned int minute = fmfromdecl(time + (sep3 + 1), sep4 - (sep3 + 1));
    unsigned int second = fmfromdecl(time + (sep4 + 1), tlen - (sep4 + 1));
    result = ((year - 1980) & 0x7f) << 25 | (month & 0x0f) << 21 | (day & 0x1f) << 16 | (hour & 0x1f) << 11 | (minute & 0x3f) << 5 | ((second >> 1) & 0x1f);
  }
  return result;
}

// inlistfile outdiskfile
int main(int argc, char const** argv) {
  unsigned char result = 0;
  char line[512];
  unsigned int tokenbounds[32];
  if (argc == 3) {
    FILE* filelist = fopen(argv[1], "rb");
    if (filelist != NULL) {
      unsigned int linelen = 512;
      unsigned int ntok;
      mmreadline(filelist, line, &linelen);
      ntok = strtokens(line, linelen, tokenbounds, 17);
      struct diskparameters dpara;
      unsigned char numtracks;
      if (ntok >= 12) {
        result = 1;
        dpara.disktype = (unsigned char)decodenuml(line + tokenbounds[1*2], tokenbounds[1*2+1]);
        numtracks = (unsigned char)decodenuml(line + tokenbounds[2*2], tokenbounds[2*2+1]);
        dpara.version = (unsigned char)decodenuml(line + tokenbounds[3*2], tokenbounds[3*2+1]);
        dpara.bytespersector = (unsigned short)decodenuml(line + tokenbounds[4*2], tokenbounds[4*2+1]);
        dpara.sectorspercluster = (unsigned char)decodenuml(line + tokenbounds[5*2], tokenbounds[5*2+1]);
        dpara.reservedsectors = (unsigned short)decodenuml(line + tokenbounds[6*2], tokenbounds[6*2+1]);
        dpara.numfats = (unsigned char)decodenuml(line + tokenbounds[7*2], tokenbounds[7*2+1]);
        dpara.maxrootdirentries = (unsigned short)decodenuml(line + tokenbounds[8*2], tokenbounds[8*2+1]);
        dpara.numsectors16 = (unsigned short)decodenuml(line + tokenbounds[9*2], tokenbounds[9*2+1]);
        dpara.mediadesc = (unsigned char)decodenuml(line + tokenbounds[10*2], tokenbounds[10*2+1]);
        dpara.sectorsperfat = (unsigned char)decodenuml(line + tokenbounds[11*2], tokenbounds[11*2+1]);
        if (dpara.version >= 30 && ntok >= 15) {
          dpara.sectorspertrack = (unsigned short)decodenuml(line + tokenbounds[12*2], tokenbounds[12*2+1]);
          dpara.numheads = (unsigned short)decodenuml(line + tokenbounds[13*2], tokenbounds[13*2+1]);
          dpara.hiddensectors = (unsigned short)decodenuml(line + tokenbounds[14*2], tokenbounds[14*2+1]);
        } else {
          dpara.sectorspertrack = 8;
          dpara.numheads = 2;
          dpara.hiddensectors = 0;
        }
      }
      char* codestub = NULL;
      unsigned int codestublen = 0;
      if (result) {
        line[tokenbounds[0] + tokenbounds[1]] = 0;
        FILE* codestubfile = fopen(line + tokenbounds[0], "rb");
        if (codestubfile) {
          fseek(codestubfile, 0, 2);
          long int fsz = ftell(codestubfile);
          codestub = malloc(fsz);
          if (codestub) {
            fseek(codestubfile, 0, 0);
            codestublen = fread(codestub, 1, fsz, codestubfile);
          } else {
            result = 0;
          }
          fclose(codestubfile);
        } else {
          result = 0;
        }
      }
      unsigned int dexdcap = 128;
      struct dextradata* fextradata = result ? malloc(sizeof(struct dextradata) * dexdcap) : NULL;
      result = (fextradata != NULL);
      struct strbuffer hostpathbuffer;
      struct fsdir diskroot;
      struct idxalloc ida;
      hostpathbuffer.strdata = NULL;
      diskroot.dirs = NULL;
      ida.array = NULL;
      result = result && strbuffinit(&hostpathbuffer, 4096, 128);
      result = result && idxinit(&ida, 128);
      result = result && fsinitdir(&diskroot, 8, 64, &ida);
      if (result) {
        while (result && !feof(filelist)) {
          linelen = 512;
          mmreadline(filelist, line, &linelen);

          ntok = strtokens(line, linelen, tokenbounds, 7);
          if (ntok < 5) {
            continue;
          }
          line[tokenbounds[0] + tokenbounds[1]] = 0;

          FILE* filetest = u8fopen(line + tokenbounds[0], "rb");
          unsigned int flen = 0;
          if (filetest != NULL) {
            fseek(filetest, 0, 2);
            flen = (unsigned int)ftell(filetest);
             fclose(filetest);
          } else {
            printf("PACKDCP in file list %s: file %s could not be opened\n", argv[1], line + tokenbounds[0]);
            continue;
          }

          unsigned int ctime = datetotstamp(line + tokenbounds[4],  tokenbounds[5], line + tokenbounds[6],  tokenbounds[7]);
          unsigned char fflags = cvtflags(line + tokenbounds[8], tokenbounds[9]);
          unsigned int hostpathidx = hostpathbuffer.entrynum;
          result = strbuffadd(&hostpathbuffer, line + tokenbounds[0], tokenbounds[1] + 1);
          line[tokenbounds[2] + tokenbounds[3]] = 0;

          unsigned short clusterofs = ntok >= 6 ? (unsigned short)decodenuml(line + tokenbounds[5*2], tokenbounds[5*2+1]) : (unsigned short)-1;

          struct fsfile* fsf = NULL;
          if (result) {
            fsf = dcpaddfile(&diskroot.filedata, line + tokenbounds[2], &ida);
            if (fsf != NULL) {
              
            } else {
              result = 0;
            }
          }
          if (result) {
            if ((fsf->fileidx >> 1) >= dexdcap) {
              struct dextradata* dexdarr2 = realloc(fextradata, sizeof(struct dextradata) * (dexdcap * 2));
              if (dexdarr2 != NULL) {
                fextradata = dexdarr2;
                dexdcap *= 2;
              } else {
                result = 0;
              }
            }
          }
          if (result) {
            struct dextradata* dexd = &fextradata[fsf->fileidx >> 1];
            dexd->flags = fflags;
            dexd->creationtime = ctime;
            dexd->clusteroffset = clusterofs;
            dexd->datalength = flen;
            dexd->hostpath = hostpathidx;
          }
        }
      }
      if (result) {
        char* reservedc = malloc(4096);
        if (reservedc != NULL) {
          memset(reservedc, 0, sizeof(char) * 4096);
          unsigned int clustersize = (unsigned int)dpara.sectorspercluster * dpara.bytespersector;
          dcpmarkreserved(&diskroot, clustersize, reservedc, fextradata);
          unsigned int maxc1 = dcpfssetoffsets(&diskroot, 2, clustersize, reservedc, 0, fextradata);
          maxc1 = dcpfssetoffsets(&diskroot, maxc1, clustersize, reservedc, 1, fextradata);
          free(reservedc);
          unsigned int dirsectors = (dpara.maxrootdirentries * 32 + dpara.bytespersector - 1) / dpara.bytespersector;
          unsigned int maxsectors = 1 + dpara.numfats * dpara.sectorsperfat + dirsectors + maxc1 * dpara.sectorspercluster;
          unsigned int fewesttracks = (maxsectors + dpara.sectorspertrack - 1) / dpara.sectorspertrack;
          if (fewesttracks > numtracks) {

            numtracks = fewesttracks;
          }
          unsigned int mintrack = numtracks > 160 ? 160 : numtracks;
          for (unsigned int idx = 0; idx < 160; ++idx) {
            dpara.tracks[idx] = (idx < mintrack);
          }
          dpara.trackflag = (mintrack == 160);
          if (mintrack < numtracks) {
            printf("File list %s has too much data for the DCP format (%d sectors vs %d max with 160 tracks).\n", argv[1], maxsectors, 160 * dpara.sectorspertrack);
            result = 0;
          }
        } else {
          result = 0;
        }
      }
      if (result) {
        FILE* outfile = fopen(argv[2], "wb");
        if (outfile != NULL) {
          dcpwritedisk(outfile, &dpara, &diskroot, &hostpathbuffer, codestub, codestublen, fextradata);
          fclose(outfile);
        } else {
          result = 0;
        }
      }
      if (diskroot.dirs != NULL) {
        fsremovefile(&diskroot.filedata, &ida);
      }
      free(ida.array);
      if (hostpathbuffer.strdata != NULL) {
        strbufffree(&hostpathbuffer);
      }
      free(codestub);
      fclose(filelist);
    }
  }
  return result != 0 ? 0 : 1;
}